"""State drift detector for runtime anomaly checks."""

from __future__ import annotations

import random
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from orchesis.engine import evaluate
from orchesis.replay import ReplayEngine, read_events_from_jsonl
from orchesis.state import DEFAULT_SESSION_ID, RateLimitTracker
from orchesis.telemetry import DecisionEvent


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


@dataclass
class DriftEvent:
    timestamp: str
    agent_id: str
    drift_type: str
    expected: str
    actual: str
    severity: str


class DriftDetector:
    """Detects state drift and runtime anomalies."""

    def __init__(self) -> None:
        self._events: list[DriftEvent] = []
        self._baseline_latency: float | None = None
        self._last_replay_sample_size: int = 0

    def _record(
        self,
        *,
        agent_id: str,
        drift_type: str,
        expected: str,
        actual: str,
        severity: str,
    ) -> DriftEvent:
        event = DriftEvent(
            timestamp=_now_iso(),
            agent_id=agent_id,
            drift_type=drift_type,
            expected=expected,
            actual=actual,
            severity=severity,
        )
        self._events.append(event)
        return event

    def check_counter_integrity(
        self,
        tracker: RateLimitTracker,
        agent_id: str,
        tool: str,
        expected_count: int,
    ) -> DriftEvent | None:
        """Verify rate limit counter matches expected value."""
        actual_count = tracker.get_count(
            tool_name=tool,
            window_seconds=86400,
            agent_id=agent_id,
            session_id=DEFAULT_SESSION_ID,
        )
        if actual_count == expected_count:
            return None
        delta = abs(actual_count - expected_count)
        severity = "medium" if delta <= 1 else "high"
        return self._record(
            agent_id=agent_id,
            drift_type="counter_mismatch",
            expected=str(expected_count),
            actual=str(actual_count),
            severity=severity,
        )

    def check_budget_integrity(
        self,
        tracker: RateLimitTracker,
        agent_id: str,
        expected_spent: float,
    ) -> DriftEvent | None:
        """Verify budget spent matches expected value."""
        actual_spent = tracker.get_agent_budget_spent(
            agent_id=agent_id,
            window_seconds=86400,
            session_id=DEFAULT_SESSION_ID,
        )
        if abs(actual_spent - expected_spent) < 1e-9:
            return None
        delta = abs(actual_spent - expected_spent)
        severity = "medium" if delta <= 0.01 else "high"
        return self._record(
            agent_id=agent_id,
            drift_type="budget_mismatch",
            expected=f"{expected_spent:.6f}",
            actual=f"{actual_spent:.6f}",
            severity=severity,
        )

    def check_replay_consistency(
        self,
        event: DecisionEvent,
        policy: dict[str, Any],
        registry=None,  # noqa: ANN001
    ) -> DriftEvent | None:
        """Replay one decision and check for divergence."""
        _ = registry  # Reserved for future identity-aware replay.
        result = ReplayEngine().replay_event(event, policy, strict=True)
        if result.match:
            return None
        drift_detail = (
            result.drift_reasons[0] if result.drift_reasons else f"original={event.decision}"
        )
        return self._record(
            agent_id=event.agent_id,
            drift_type="replay_divergence",
            expected=event.decision,
            actual=drift_detail,
            severity="critical",
        )

    def check_latency_anomaly(self, duration_us: int) -> DriftEvent | None:
        """Detect latency spikes (>10x baseline)."""
        if duration_us <= 0:
            return None
        if self._baseline_latency is None:
            self._baseline_latency = float(duration_us)
            return None
        baseline = max(self._baseline_latency, 1.0)
        if duration_us > baseline * 10.0:
            return self._record(
                agent_id="__runtime__",
                drift_type="latency_spike",
                expected=f"<= {baseline * 10.0:.0f}us",
                actual=f"{duration_us}us",
                severity="medium",
            )
        self._baseline_latency = (baseline * 0.95) + (float(duration_us) * 0.05)
        return None

    def check_decision_consistency(
        self,
        request: dict[str, Any],
        policy: dict[str, Any],
        expected_allowed: bool,
        registry=None,  # noqa: ANN001
    ) -> DriftEvent | None:
        """Verify a known request produces expected decision."""
        decision = evaluate(
            request,
            policy,
            state=RateLimitTracker(persist_path=None),
            registry=registry,
        )
        if decision.allowed == expected_allowed:
            return None
        drift_type = "unexpected_allow" if decision.allowed else "unexpected_deny"
        expected = "ALLOW" if expected_allowed else "DENY"
        actual = "ALLOW" if decision.allowed else "DENY"
        severity = "critical" if drift_type == "unexpected_allow" else "high"
        context = request.get("context")
        agent_id = (
            context.get("agent", "__global__") if isinstance(context, dict) else "__global__"
        )
        return self._record(
            agent_id=str(agent_id),
            drift_type=drift_type,
            expected=expected,
            actual=actual,
            severity=severity,
        )

    def run_all_checks(
        self,
        tracker: RateLimitTracker,
        policy: dict[str, Any],
        decisions_log: str | Path,
        registry=None,  # noqa: ANN001
    ) -> list[DriftEvent]:
        """Run all drift checks. Returns list of drift events."""
        _ = tracker
        path = Path(decisions_log)
        events = read_events_from_jsonl(path)
        if not events:
            self._last_replay_sample_size = 0
            return list(self._events)

        rng = random.Random(42)
        sample_size = min(100, len(events))
        self._last_replay_sample_size = sample_size
        sampled = [events[idx] for idx in rng.sample(range(len(events)), sample_size)]
        for item in sampled:
            self.check_latency_anomaly(item.evaluation_duration_us)
            self.check_replay_consistency(item, policy, registry=registry)
        return list(self._events)

    @property
    def baseline_latency_us(self) -> float | None:
        return self._baseline_latency

    @property
    def replay_sample_size(self) -> int:
        return self._last_replay_sample_size

    @property
    def events(self) -> list[DriftEvent]:
        return list(self._events)

    @property
    def has_critical_drift(self) -> bool:
        return any(event.severity == "critical" for event in self._events)



from __future__ import annotations

from orchesis.ahocorasick import AhoCorasickMatcher


def test_single_pattern_found() -> None:
    matcher = AhoCorasickMatcher({"p1": "secret"})
    matches = matcher.search("my secret here")
    assert len(matches) == 1
    assert matches[0].pattern_id == "p1"


def test_single_pattern_not_found() -> None:
    matcher = AhoCorasickMatcher({"p1": "secret"})
    assert matcher.search("clean text") == []


def test_multiple_patterns_all_found() -> None:
    matcher = AhoCorasickMatcher({"a": "alpha", "b": "beta", "g": "gamma"})
    text = "alpha beta gamma"
    ids = [item.pattern_id for item in matcher.search(text)]
    assert {"a", "b", "g"}.issubset(set(ids))


def test_multiple_patterns_some_found() -> None:
    matcher = AhoCorasickMatcher({"a": "alpha", "b": "beta"})
    ids = [item.pattern_id for item in matcher.search("alpha only")]
    assert "a" in ids
    assert "b" not in ids


def test_overlapping_matches_he_her() -> None:
    matcher = AhoCorasickMatcher({"he": "he", "her": "her"})
    matches = matcher.search("her")
    found = {(item.pattern_id, item.start, item.end) for item in matches}
    assert ("he", 0, 2) in found
    assert ("her", 0, 3) in found


def test_empty_text() -> None:
    matcher = AhoCorasickMatcher({"a": "x"})
    assert matcher.search("") == []


def test_empty_patterns() -> None:
    matcher = AhoCorasickMatcher({})
    assert matcher.search("anything") == []


def test_unicode_patterns_cyrillic() -> None:
    matcher = AhoCorasickMatcher({"ru": "привет"})
    matches = matcher.search("мир привет мир")
    assert len(matches) == 1
    assert matches[0].matched_text == "привет"


def test_unicode_patterns_cjk() -> None:
    matcher = AhoCorasickMatcher({"cjk": "安全"})
    matches = matcher.search("这是安全测试")
    assert matches and matches[0].pattern_id == "cjk"


def test_unicode_patterns_emoji() -> None:
    matcher = AhoCorasickMatcher({"em": "🔒"})
    matches = matcher.search("token 🔒 hidden")
    assert len(matches) == 1


def test_case_insensitive_matching() -> None:
    matcher = AhoCorasickMatcher({"a": "Secret"}, case_insensitive=True)
    matches = matcher.search("my secret")
    assert len(matches) == 1
    assert matches[0].pattern_id == "a"


def test_very_long_text() -> None:
    matcher = AhoCorasickMatcher({"x": "needle"})
    text = ("a" * 100_000) + "needle"
    matches = matcher.search(text)
    assert len(matches) == 1
    assert matches[0].start == 100_000


def test_pattern_start_middle_end() -> None:
    matcher = AhoCorasickMatcher({"s": "start", "m": "middle", "e": "end"})
    matches = matcher.search("start and middle and end")
    ids = [m.pattern_id for m in matches]
    assert "s" in ids and "m" in ids and "e" in ids


def test_search_first_returns_first_match() -> None:
    matcher = AhoCorasickMatcher({"a": "alpha", "b": "beta"})
    first = matcher.search_first("zz alpha then beta")
    assert first is not None
    assert first.pattern_id == "a"


def test_duplicate_patterns_different_ids() -> None:
    matcher = AhoCorasickMatcher({"a1": "key", "a2": "key"})
    matches = matcher.search("key")
    ids = sorted(item.pattern_id for item in matches)
    assert ids == ["a1", "a2"]



"""Dedicated MAST failure mode detectors (H2.3)."""

from __future__ import annotations

from collections import Counter, defaultdict, deque
from dataclasses import asdict, dataclass
import json
import math
import re
import threading
import time
from typing import Any, Optional

MODEL_CONTEXT_WINDOWS = {
    "gpt-4o": 128_000,
    "gpt-4o-mini": 128_000,
    "gpt-4-turbo": 128_000,
    "gpt-4": 8_192,
    "gpt-3.5-turbo": 16_385,
    "o1": 200_000,
    "o1-mini": 128_000,
    "o3": 200_000,
    "o3-mini": 200_000,
    "claude-sonnet-4-20250514": 200_000,
    "claude-opus-4-20250514": 200_000,
    "claude-3-5-sonnet-20241022": 200_000,
    "claude-3-5-haiku-20241022": 200_000,
    "gemini-2.0-flash": 1_000_000,
    "gemini-2.0-pro": 2_000_000,
    "gemini-1.5-pro": 2_000_000,
    "gemini-1.5-flash": 1_000_000,
    "_default": 128_000,
}

_SEVERITY_SCORE = {"low": 5, "medium": 12, "high": 25, "critical": 40}
_GRADE_VALUE = {"A": 5, "B": 4, "C": 3, "D": 2, "F": 1}
_SHELL_RE = re.compile(r"(?:\brm\s+-rf\b|\bcurl\b.+\|\s*(?:bash|sh)|\bwget\b.+\|\s*(?:bash|sh)|&&|;)")
_SQLI_RE = re.compile(r"(?:\bunion\s+select\b|\bdrop\s+table\b|'\s*or\s*1=1)", re.IGNORECASE)
_PATH_RE = re.compile(r"(?:\.\./|\.\.\\|/etc/passwd|\\windows\\system32)", re.IGNORECASE)
_SSRF_RE = re.compile(
    r"https?://(?:127\.0\.0\.1|localhost|169\.254\.169\.254|10\.\d+\.\d+\.\d+|192\.168\.\d+\.\d+|172\.(?:1[6-9]|2\d|3[0-1])\.\d+\.\d+)",
    re.IGNORECASE,
)
_API_KEY_RE = re.compile(r"\b(?:sk-[A-Za-z0-9_\-]{10,}|ghp_[A-Za-z0-9]{20,}|AKIA[0-9A-Z]{16})\b")
_JWT_RE = re.compile(r"\beyJ[A-Za-z0-9_-]{8,}\.[A-Za-z0-9_-]{8,}\.[A-Za-z0-9_-]{8,}\b")
_CONN_RE = re.compile(r"\b(?:postgres(?:ql)?|mongodb|redis)://[^\s]+", re.IGNORECASE)
_PRIVATE_KEY_RE = re.compile(r"-----BEGIN (?:RSA |EC |OPENSSH )?PRIVATE KEY-----")
_SOCIAL_RE = re.compile(
    r"(?:click here|enter your password|run this command|download this file|disable security)",
    re.IGNORECASE,
)
_URL_RE = re.compile(r"https?://[^\s)>\]]+", re.IGNORECASE)
_HIDDEN_UNICODE_RE = re.compile(r"[\u200b-\u200f\u2060\ufeff]")
_B64_RE = re.compile(r"(?:[A-Za-z0-9+/]{200,}={0,2})")


@dataclass
class MASTFinding:
    """A single MAST failure mode detection."""

    failure_mode: str
    name: str
    severity: str
    confidence: float
    description: str
    evidence: dict[str, Any]
    recommendation: str
    timestamp: float


class MASTDetectors:
    """Dedicated detectors for selected MAST failure modes."""

    def __init__(self, config: Optional[dict] = None):
        cfg = config if isinstance(config, dict) else {}
        self.enabled = bool(cfg.get("enabled", True))
        det_cfg = cfg.get("detectors") if isinstance(cfg.get("detectors"), dict) else {}
        self._enabled = {
            "privilege_escalation": bool(det_cfg.get("privilege_escalation", True)),
            "tool_abuse": bool(det_cfg.get("tool_abuse", True)),
            "credential_leakage": bool(det_cfg.get("credential_leakage", True)),
            "context_overflow": bool(det_cfg.get("context_overflow", True)),
            "cascading_failure": bool(det_cfg.get("cascading_failure", True)),
            "output_manipulation": bool(det_cfg.get("output_manipulation", True)),
            "observability_gap": bool(det_cfg.get("observability_gap", True)),
            "compliance_drift": bool(det_cfg.get("compliance_drift", True)),
        }
        self._lock = threading.Lock()
        self._agent_findings: dict[str, list[MASTFinding]] = defaultdict(list)
        self._agent_last_seen: dict[str, float] = {}
        self._tool_history: dict[str, deque[tuple[float, str, str]]] = defaultdict(lambda: deque(maxlen=500))
        self._error_events: deque[tuple[float, str, str]] = deque(maxlen=2000)
        self._stats = {
            "checks_request": 0,
            "checks_response": 0,
            "findings_total": 0,
            "findings_by_mode": defaultdict(int),
            "findings_by_severity": defaultdict(int),
        }

    @staticmethod
    def _tokens(text: str) -> int:
        return len((text or "").split())

    @staticmethod
    def _extract_messages(request_data: dict[str, Any]) -> list[dict[str, Any]]:
        messages = request_data.get("messages")
        return messages if isinstance(messages, list) else []

    @staticmethod
    def _extract_tool_calls(request_data: dict[str, Any]) -> list[dict[str, Any]]:
        out: list[dict[str, Any]] = []
        for msg in MASTDetectors._extract_messages(request_data):
            if not isinstance(msg, dict):
                continue
            calls = msg.get("tool_calls")
            if isinstance(calls, list):
                for call in calls:
                    if isinstance(call, dict):
                        name = (
                            call.get("name")
                            or (call.get("function") or {}).get("name")
                            or call.get("tool_name")
                            or ""
                        )
                        args = (
                            call.get("arguments")
                            or call.get("params")
                            or call.get("input")
                            or (call.get("function") or {}).get("arguments")
                            or ""
                        )
                        out.append({"name": str(name), "args": args})
        req_calls = request_data.get("tool_calls")
        if isinstance(req_calls, list):
            for call in req_calls:
                if isinstance(call, dict):
                    out.append({"name": str(call.get("name", "")), "args": call.get("args", call.get("params", {}))})
        return out

    @staticmethod
    def _context_window(model_name: str) -> int:
        return int(MODEL_CONTEXT_WINDOWS.get(model_name, MODEL_CONTEXT_WINDOWS["_default"]))

    @staticmethod
    def _char_entropy(text: str) -> float:
        if not text:
            return 0.0
        freq = Counter(text)
        total = float(len(text))
        score = 0.0
        for count in freq.values():
            p = count / total
            if p > 0.0:
                score -= p * math.log2(p)
        return score

    @staticmethod
    def _mk(
        mode: str,
        name: str,
        severity: str,
        confidence: float,
        desc: str,
        evidence: dict[str, Any],
        recommendation: str,
    ) -> MASTFinding:
        return MASTFinding(
            failure_mode=mode,
            name=name,
            severity=severity,
            confidence=max(0.0, min(1.0, float(confidence))),
            description=desc,
            evidence=evidence,
            recommendation=recommendation,
            timestamp=time.time(),
        )

    def _record(self, agent_id: str, finding: MASTFinding) -> None:
        with self._lock:
            self._agent_findings[agent_id].append(finding)
            self._agent_last_seen[agent_id] = finding.timestamp
            self._stats["findings_total"] += 1
            self._stats["findings_by_mode"][finding.failure_mode] += 1
            self._stats["findings_by_severity"][finding.severity] += 1

    def _check_privilege_escalation(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        _ = agent_id
        approved_tools = context.get("approved_tools")
        approved_tools_set = set(approved_tools) if isinstance(approved_tools, list) else set()
        approved_models = context.get("approved_models")
        approved_models_set = set(approved_models) if isinstance(approved_models, list) else set()
        messages = self._extract_messages(request_data)
        model = str(request_data.get("model", "") or "")
        calls = self._extract_tool_calls(request_data)
        for msg in messages:
            if isinstance(msg, dict) and str(msg.get("role", "")).lower() == "system":
                return self._mk(
                    "FM-1.3",
                    "Privilege Escalation",
                    "critical",
                    0.98,
                    "Agent attempted role escalation to system.",
                    {"role": "system"},
                    "Reject system-role injection and isolate session.",
                )
        for msg in messages:
            text = str(msg.get("content", "") if isinstance(msg, dict) else "")
            if "ignore previous instructions" in text.lower() or "override system" in text.lower():
                return self._mk(
                    "FM-1.3",
                    "Privilege Escalation",
                    "high",
                    0.9,
                    "Potential system prompt modification attempt detected.",
                    {"snippet": text[:120]},
                    "Block instruction override and enforce immutable system policy.",
                )
        if approved_models_set and model and model not in approved_models_set:
            return self._mk(
                "FM-1.3",
                "Privilege Escalation",
                "medium",
                0.8,
                "Agent requested unapproved model.",
                {"model": model},
                "Restrict model routing to approved model set.",
            )
        if approved_tools_set:
            for call in calls:
                name = str(call.get("name", ""))
                if name and name not in approved_tools_set:
                    return self._mk(
                        "FM-1.3",
                        "Privilege Escalation",
                        "medium",
                        0.78,
                        "Agent requested unapproved tool.",
                        {"tool": name},
                        "Enforce capabilities.allowlist and reject unapproved tool.",
                    )
        budget = context.get("token_budget", {})
        if isinstance(budget, dict):
            max_allowed = int(budget.get("max_tokens", 0) or 0)
            requested = int(request_data.get("max_completion_tokens", request_data.get("max_tokens", 0)) or 0)
            if max_allowed > 0 and requested > max_allowed:
                return self._mk(
                    "FM-1.3",
                    "Privilege Escalation",
                    "high",
                    0.85,
                    "Agent attempted token limit override beyond policy budget.",
                    {"requested": requested, "allowed": max_allowed},
                    "Clamp max tokens to policy budget and alert operator.",
                )
        return None

    def _check_tool_abuse(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        _ = context
        calls = self._extract_tool_calls(request_data)
        now = time.time()
        if not calls:
            return None
        history = self._tool_history[agent_id]
        burst_count = 0
        identical_counter: Counter[tuple[str, str]] = Counter()
        for call in calls:
            name = str(call.get("name", "") or "")
            raw_args = call.get("args", {})
            args_str = jsonable(raw_args)
            history.append((now, name, args_str))
            if now - history[-1][0] <= 60.0:
                burst_count += 1
            identical_counter[(name, args_str)] += 1
            if not args_str or args_str in {"{}", "[]", '""'}:
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "low",
                    0.6,
                    "Tool call has empty or malformed arguments.",
                    {"tool": name},
                    "Validate required tool arguments before execution.",
                )
            lowered = args_str.lower()
            if _SHELL_RE.search(lowered):
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "critical",
                    0.95,
                    "Shell injection-like pattern in tool arguments.",
                    {"tool": name, "args": args_str[:120]},
                    "Block execution and quarantine request for review.",
                )
            if _SQLI_RE.search(lowered):
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "critical",
                    0.95,
                    "SQL injection-like pattern in tool arguments.",
                    {"tool": name},
                    "Deny request and sanitize SQL inputs with parameterization.",
                )
            if _SSRF_RE.search(lowered):
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "critical",
                    0.92,
                    "Potential SSRF/internal-network target in tool arguments.",
                    {"tool": name},
                    "Block internal URL access and enforce allowlisted domains.",
                )
            if _PATH_RE.search(lowered):
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "high",
                    0.88,
                    "Path traversal/sensitive path pattern detected.",
                    {"tool": name},
                    "Restrict filesystem scope and sanitize path arguments.",
                )
        if any(v >= 3 for v in identical_counter.values()):
            tool, _args = max(identical_counter.items(), key=lambda kv: kv[1])[0]
            return self._mk(
                "FM-1.4",
                "Tool Abuse",
                "medium",
                0.8,
                "Repeated identical tool calls detected (stuck agent pattern).",
                {"tool": tool},
                "Add de-duplication/backoff for repeated tool invocations.",
            )
        recent = [item for item in history if now - item[0] <= 60.0]
        if len(recent) >= 10:
            baseline = max(1, len(history) - len(recent))
            if len(recent) >= baseline * 10:
                return self._mk(
                    "FM-1.4",
                    "Tool Abuse",
                    "medium",
                    0.75,
                    "Tool call frequency spike detected for this agent.",
                    {"recent_calls_60s": len(recent), "baseline_calls": baseline},
                    "Throttle tool calls and apply per-tool rate limits.",
                )
        return None

    def _check_credential_leakage(self, agent_id: str, response_data: dict[str, Any], original_request: dict[str, Any]) -> Optional[MASTFinding]:
        _ = (agent_id, original_request)
        content = response_text(response_data)
        if not content:
            return None
        if _PRIVATE_KEY_RE.search(content) or _JWT_RE.search(content):
            return self._mk(
                "FM-1.5",
                "Credential Leakage",
                "critical",
                0.98,
                "High-confidence credential material detected in response.",
                {"kind": "private_key_or_jwt"},
                "Redact response content and rotate potentially leaked credentials.",
            )
        if _API_KEY_RE.search(content) or _CONN_RE.search(content):
            return self._mk(
                "FM-1.5",
                "Credential Leakage",
                "high",
                0.9,
                "Potential API key or credential string found in response output.",
                {"kind": "api_key_or_connection_string"},
                "Block response and scrub sensitive data before returning to user.",
            )
        for token in re.findall(r"[A-Za-z0-9_\-+/=]{20,}", content):
            if self._char_entropy(token) > 4.5 and len(token) > 20:
                return self._mk(
                    "FM-1.5",
                    "Credential Leakage",
                    "high",
                    0.75,
                    "High-entropy token-like value found in response output.",
                    {"sample_len": len(token)},
                    "Inspect output and redact suspicious secret-like values.",
                )
        return None

    def _check_context_overflow(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        _ = agent_id
        model = str(request_data.get("model", "") or "")
        window = self._context_window(model)
        messages = self._extract_messages(request_data)
        lengths = [self._tokens(str(msg.get("content", "") if isinstance(msg, dict) else "")) for msg in messages]
        total = sum(lengths)
        if lengths and max(lengths) > int(window * 0.5):
            return self._mk(
                "FM-2.3",
                "Context Window Overflow",
                "critical",
                0.95,
                "Single message consumes >50% of model context window.",
                {"max_message_tokens": max(lengths), "context_window": window},
                "Reject oversized message and require chunking/summarization.",
            )
        usage_ratio = total / float(max(1, window))
        if usage_ratio > 0.8:
            return self._mk(
                "FM-2.3",
                "Context Window Overflow",
                "high",
                0.85,
                "Total prompt size exceeds 80% of context window.",
                {"total_tokens": total, "context_window": window},
                "Apply context compression and preserve core instructions.",
            )
        if usage_ratio > 0.6:
            return self._mk(
                "FM-2.3",
                "Context Window Overflow",
                "medium",
                0.65,
                "Prompt is approaching context limit.",
                {"ratio": round(usage_ratio, 3)},
                "Enable proactive trimming/summarization before overflow.",
            )
        for msg in messages:
            text = str(msg.get("content", "") if isinstance(msg, dict) else "")
            if _B64_RE.search(text):
                return self._mk(
                    "FM-2.3",
                    "Context Window Overflow",
                    "high",
                    0.8,
                    "Large encoded blob detected (possible context-filling attack).",
                    {"model": model or "_default"},
                    "Reject large encoded payloads and request structured input.",
                )
        if len(lengths) >= 4:
            grows = sum(1 for i in range(1, len(lengths)) if lengths[i] > lengths[i - 1] * 1.5 and lengths[i] > 200)
            if grows >= 3:
                return self._mk(
                    "FM-2.3",
                    "Context Window Overflow",
                    "high",
                    0.78,
                    "Rapid message growth trend indicates potential overflow strategy.",
                    {"growth_events": grows},
                    "Rate-limit large turns and enforce progressive size caps.",
                )
        return None

    def _check_cascading_failure(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        now = time.time()
        failed = bool(context.get("request_failed", False))
        status_code = int(request_data.get("status_code", 0) or 0)
        if status_code >= 500:
            failed = True
        if failed:
            self._error_events.append((now, agent_id, "error"))
        retry_count = int(context.get("retry_count", 0) or 0)
        if retry_count > 2:
            self._error_events.append((now, agent_id, "retry"))
        recent = [item for item in self._error_events if now - item[0] <= 60.0]
        failed_agents = {a for ts, a, kind in recent if kind == "error" and now - ts <= 60.0}
        if len(failed_agents) >= 3:
            return self._mk(
                "FM-2.6",
                "Cascading Failure",
                "high" if len(failed_agents) < 4 else "critical",
                0.9 if len(failed_agents) >= 4 else 0.82,
                "Multiple agents failing concurrently indicates cascading behavior.",
                {"failed_agents_60s": len(failed_agents)},
                "Open circuit for shared dependencies and isolate failing workloads.",
            )
        if len(failed_agents) == 2:
            return self._mk(
                "FM-2.6",
                "Cascading Failure",
                "medium",
                0.7,
                "Two agents failed within the same short time window.",
                {"failed_agents_60s": 2},
                "Monitor shared dependencies and apply backpressure.",
            )
        retries = [item for item in recent if item[2] == "retry"]
        if len(retries) >= 5:
            return self._mk(
                "FM-2.6",
                "Cascading Failure",
                "high",
                0.8,
                "Retry storm detected across agents.",
                {"retries_60s": len(retries)},
                "Enable exponential backoff and enforce retry budgets.",
            )
        return None

    def _check_output_manipulation(self, agent_id: str, response_data: dict[str, Any], original_request: dict[str, Any]) -> Optional[MASTFinding]:
        _ = agent_id
        role = str(response_data.get("role", "") or "")
        if role and role.lower() != "assistant":
            return self._mk(
                "FM-3.1",
                "Output Manipulation",
                "critical",
                0.95,
                "Response role mismatch detected.",
                {"role": role},
                "Reject response and verify upstream response integrity.",
            )
        text = response_text(response_data)
        if not text:
            return None
        if _HIDDEN_UNICODE_RE.search(text):
            return self._mk(
                "FM-3.1",
                "Output Manipulation",
                "critical",
                0.95,
                "Hidden/invisible Unicode content detected in output.",
                {"contains_hidden_unicode": True},
                "Strip invisible characters and alert operator.",
            )
        if _SOCIAL_RE.search(text):
            return self._mk(
                "FM-3.1",
                "Output Manipulation",
                "high",
                0.86,
                "Potential social-engineering instructions found in output.",
                {"snippet": text[:120]},
                "Warn user and require explicit confirmation before actions.",
            )
        urls = _URL_RE.findall(text)
        if urls:
            req_text = response_text(original_request or {})
            if not any(url in req_text for url in urls):
                return self._mk(
                    "FM-3.1",
                    "Output Manipulation",
                    "medium",
                    0.7,
                    "Response contains URLs not present in original request context.",
                    {"urls": urls[:3]},
                    "Flag suspicious links and enforce URL allowlist policy.",
                )
        return None

    def _check_observability_gap(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        now = time.time()
        session_id = str(request_data.get("session_id", context.get("session_id", "")) or "")
        model = str(request_data.get("model", "") or "")
        if not session_id and not str(request_data.get("request_id", "") or ""):
            return self._mk(
                "OE-1",
                "Observability Gap",
                "critical",
                0.92,
                "Request is completely untracked (missing session and request identifiers).",
                {},
                "Inject mandatory IDs at ingress and reject untracked traffic.",
            )
        if not session_id:
            return self._mk(
                "OE-1",
                "Observability Gap",
                "medium",
                0.7,
                "Missing session identifier reduces traceability.",
                {},
                "Ensure session id is propagated across all agent requests.",
            )
        if not model:
            return self._mk(
                "OE-1",
                "Observability Gap",
                "low",
                0.55,
                "Model field missing; cost and attribution may be incomplete.",
                {},
                "Populate model metadata before routing.",
            )
        with self._lock:
            last = self._agent_last_seen.get(agent_id)
            self._agent_last_seen[agent_id] = now
        if isinstance(last, float) and (now - last) > float(context.get("max_gap_seconds", 600.0)):
            return self._mk(
                "OE-1",
                "Observability Gap",
                "high",
                0.8,
                "Large unexplained activity gap detected for tracked agent.",
                {"gap_seconds": round(now - last, 3)},
                "Audit ingestion path for dropped telemetry events.",
            )
        calls = self._extract_tool_calls(request_data)
        if calls and not bool(context.get("tool_metadata_present", True)):
            return self._mk(
                "OE-1",
                "Observability Gap",
                "medium",
                0.68,
                "Tool calls present but tool metadata missing from observability context.",
                {"tool_calls": len(calls)},
                "Attach tool usage metadata in telemetry pipeline.",
            )
        return None

    def _check_compliance_drift(self, agent_id: str, request_data: dict[str, Any], context: dict[str, Any]) -> Optional[MASTFinding]:
        _ = agent_id
        model = str(request_data.get("model", "") or "")
        approved_models = context.get("approved_models")
        approved_models_set = set(approved_models) if isinstance(approved_models, list) else set()
        findings: list[tuple[str, float, str, dict[str, Any], str]] = []
        if approved_models_set and model and model not in approved_models_set:
            findings.append(
                (
                    "high",
                    0.85,
                    "Model outside approved compliance boundary.",
                    {"model": model},
                    "Route to approved model set only.",
                )
            )
        ars_hist = context.get("ars_history")
        if isinstance(ars_hist, list) and len(ars_hist) >= 2:
            try:
                first = _GRADE_VALUE.get(str(ars_hist[0]).upper(), 0)
                last = _GRADE_VALUE.get(str(ars_hist[-1]).upper(), 0)
                if first - last >= 2:
                    findings.append(
                        (
                            "high",
                            0.82,
                            "ARS grade declined by 2+ levels.",
                            {"from": str(ars_hist[0]), "to": str(ars_hist[-1])},
                            "Investigate regression and apply stricter runtime controls.",
                        )
                    )
            except Exception:
                pass
        risk_hist = context.get("risk_history")
        if isinstance(risk_hist, list) and len(risk_hist) >= 3:
            try:
                last3 = [float(x) for x in risk_hist[-3:]]
                if last3[0] < last3[1] < last3[2]:
                    sev = "critical" if last3[-1] >= 80 else "medium"
                    findings.append(
                        (
                            sev,
                            0.9 if sev == "critical" else 0.7,
                            "Session risk trending upward over time.",
                            {"recent_risk": last3},
                            "Trigger adaptive throttling and review policy drift.",
                        )
                    )
            except Exception:
                pass
        token_hist = context.get("token_usage_history")
        token_budget = context.get("token_budget")
        if isinstance(token_hist, list) and isinstance(token_budget, dict) and token_hist:
            budget_limit = float(token_budget.get("daily", token_budget.get("max_tokens", 0)) or 0.0)
            if budget_limit > 0:
                avg_recent = sum(float(x) for x in token_hist[-5:]) / float(min(5, len(token_hist)))
                if avg_recent > budget_limit * 0.5:
                    findings.append(
                        (
                            "medium",
                            0.66,
                            "Token usage trend is approaching budget ceiling.",
                            {"avg_recent": avg_recent, "budget_limit": budget_limit},
                            "Enable stronger context compression and budget controls.",
                        )
                    )
        if not findings:
            return None
        order = {"low": 0, "medium": 1, "high": 2, "critical": 3}
        severity, conf, desc, evidence, rec = max(findings, key=lambda item: order[item[0]])
        return self._mk("OE-6", "Compliance Drift", severity, conf, desc, evidence, rec)

    def check_request(self, agent_id: str, request_data: dict[str, Any], context: dict | None = None) -> list[MASTFinding]:
        if not self.enabled:
            return []
        ctx = context if isinstance(context, dict) else {}
        req = request_data if isinstance(request_data, dict) else {}
        findings: list[MASTFinding] = []
        with self._lock:
            self._stats["checks_request"] += 1
        checks = [
            ("privilege_escalation", self._check_privilege_escalation),
            ("tool_abuse", self._check_tool_abuse),
            ("context_overflow", self._check_context_overflow),
            ("cascading_failure", self._check_cascading_failure),
            ("observability_gap", self._check_observability_gap),
            ("compliance_drift", self._check_compliance_drift),
        ]
        for key, fn in checks:
            if not self._enabled.get(key, False):
                continue
            try:
                finding = fn(agent_id, req, ctx)
                if isinstance(finding, MASTFinding):
                    findings.append(finding)
                    self._record(agent_id, finding)
            except Exception:
                continue
        return findings

    def check_response(
        self,
        agent_id: str,
        response_data: dict[str, Any],
        original_request: dict | None = None,
    ) -> list[MASTFinding]:
        if not self.enabled:
            return []
        findings: list[MASTFinding] = []
        with self._lock:
            self._stats["checks_response"] += 1
        checks = [
            ("credential_leakage", self._check_credential_leakage),
            ("output_manipulation", self._check_output_manipulation),
        ]
        for key, fn in checks:
            if not self._enabled.get(key, False):
                continue
            try:
                finding = fn(agent_id, response_data if isinstance(response_data, dict) else {}, original_request or {})
                if isinstance(finding, MASTFinding):
                    findings.append(finding)
                    self._record(agent_id, finding)
            except Exception:
                continue
        return findings

    def get_agent_compliance(self, agent_id: str) -> dict[str, Any]:
        with self._lock:
            findings = list(self._agent_findings.get(agent_id, []))
            last_check = float(self._agent_last_seen.get(agent_id, 0.0))
        by_sev = {"low": 0, "medium": 0, "high": 0, "critical": 0}
        modes: set[str] = set()
        penalty = 0
        for item in findings:
            by_sev[item.severity] = by_sev.get(item.severity, 0) + 1
            modes.add(item.failure_mode)
            penalty += _SEVERITY_SCORE.get(item.severity, 5)
        score = max(0, 100 - penalty)
        return {
            "agent_id": agent_id,
            "findings_total": len(findings),
            "findings_by_severity": by_sev,
            "failure_modes_triggered": sorted(modes),
            "compliance_score": score,
            "last_check": last_check,
        }

    def get_stats(self) -> dict[str, Any]:
        with self._lock:
            return {
                "enabled": self.enabled,
                "checks_request": int(self._stats["checks_request"]),
                "checks_response": int(self._stats["checks_response"]),
                "findings_total": int(self._stats["findings_total"]),
                "findings_by_mode": dict(self._stats["findings_by_mode"]),
                "findings_by_severity": dict(self._stats["findings_by_severity"]),
                "agents_tracked": len(self._agent_findings),
            }

    def reset(self, agent_id: str) -> None:
        with self._lock:
            self._agent_findings.pop(agent_id, None)
            self._agent_last_seen.pop(agent_id, None)
            self._tool_history.pop(agent_id, None)


def jsonable(value: Any) -> str:
    try:
        return json.dumps(value, ensure_ascii=False, sort_keys=True, separators=(",", ":"))
    except Exception:
        return str(value)


def response_text(payload: dict[str, Any]) -> str:
    if not isinstance(payload, dict):
        return ""
    if isinstance(payload.get("content"), str):
        return str(payload.get("content", ""))
    choices = payload.get("choices")
    if isinstance(choices, list) and choices:
        first = choices[0]
        if isinstance(first, dict):
            msg = first.get("message")
            if isinstance(msg, dict) and isinstance(msg.get("content"), str):
                return str(msg.get("content", ""))
    output = payload.get("output_text")
    if isinstance(output, str):
        return output
    return ""



"""Agent Reliability Score (ARS) - unified 0-100 reliability metric."""

from __future__ import annotations

import math
import threading
import time
from dataclasses import dataclass, field
from typing import Any


@dataclass
class AgentMetrics:
    """Raw metrics collected per agent."""

    agent_id: str
    total_sessions: int = 0
    successful_sessions: int = 0
    failed_sessions: int = 0
    total_requests: int = 0
    loop_flagged_requests: int = 0
    total_cost_usd: float = 0.0
    budget_ceiling_usd: float = 50.0
    latencies_ms: list[float] = field(default_factory=list)
    token_counts: list[int] = field(default_factory=list)
    clean_terminations: int = 0
    dirty_terminations: int = 0
    requests_with_threats: int = 0
    first_seen: float = 0.0
    last_seen: float = 0.0


@dataclass
class ARSResult:
    """Computed ARS for an agent."""

    agent_id: str
    score: float
    grade: str
    components: dict[str, float]
    sample_size: int
    confidence: str
    computed_at: float = 0.0


class AgentReliabilityScore:
    """Compute and track ARS for all agents."""

    DEFAULT_WEIGHTS = {
        "task_completion": 20,
        "loop_freedom": 15,
        "cost_efficiency": 15,
        "latency_health": 10,
        "context_stability": 10,
        "termination_quality": 15,
        "security_compliance": 15,
    }

    def __init__(
        self,
        weights: dict[str, int] | None = None,
        latency_baseline_ms: float = 2000.0,
        max_latencies_stored: int = 1000,
        max_token_counts_stored: int = 1000,
        max_agents: int = 10_000,
        enabled: bool = True,
    ) -> None:
        self._weights = self.DEFAULT_WEIGHTS.copy()
        if isinstance(weights, dict):
            for key, value in weights.items():
                if key in self._weights and isinstance(value, int | float):
                    self._weights[key] = int(value)
        self._latency_baseline = max(100.0, float(latency_baseline_ms))
        self._max_latencies = max(100, int(max_latencies_stored))
        self._max_tokens = max(100, int(max_token_counts_stored))
        self._max_agents = max(1, int(max_agents))
        self._enabled = bool(enabled)
        self._lock = threading.Lock()
        self._agents: dict[str, AgentMetrics] = {}
        self._stats = {"total_updates": 0, "total_computes": 0, "agents_tracked": 0, "evictions": 0}

    @property
    def enabled(self) -> bool:
        return self._enabled

    def update(
        self,
        agent_id: str,
        *,
        is_session_end: bool = False,
        session_success: bool | None = None,
        loop_flagged: bool = False,
        cost_usd: float = 0.0,
        latency_ms: float = 0.0,
        token_count: int = 0,
        clean_termination: bool | None = None,
        has_threat: bool = False,
    ) -> None:
        if not self._enabled or not agent_id:
            return
        now = time.time()
        with self._lock:
            metrics = self._agents.get(agent_id)
            if metrics is None:
                if len(self._agents) >= self._max_agents:
                    oldest_agent_id = min(
                        self._agents,
                        key=lambda aid: float(self._agents[aid].last_seen or self._agents[aid].first_seen or 0.0),
                    )
                    self._agents.pop(oldest_agent_id, None)
                    self._stats["evictions"] += 1
                metrics = AgentMetrics(agent_id=agent_id, first_seen=now, last_seen=now)
                self._agents[agent_id] = metrics
                self._stats["agents_tracked"] += 1
            metrics.last_seen = now
            metrics.total_requests += 1
            self._stats["total_updates"] += 1

            if is_session_end and session_success is not None:
                metrics.total_sessions += 1
                if bool(session_success):
                    metrics.successful_sessions += 1
                else:
                    metrics.failed_sessions += 1

            if loop_flagged:
                metrics.loop_flagged_requests += 1

            metrics.total_cost_usd += float(max(0.0, cost_usd))

            if latency_ms > 0:
                metrics.latencies_ms.append(float(latency_ms))
                if len(metrics.latencies_ms) > self._max_latencies:
                    metrics.latencies_ms = metrics.latencies_ms[-self._max_latencies :]

            if token_count > 0:
                metrics.token_counts.append(int(token_count))
                if len(metrics.token_counts) > self._max_tokens:
                    metrics.token_counts = metrics.token_counts[-self._max_tokens :]

            if clean_termination is not None:
                if clean_termination:
                    metrics.clean_terminations += 1
                else:
                    metrics.dirty_terminations += 1

            if has_threat:
                metrics.requests_with_threats += 1

    def compute(self, agent_id: str) -> ARSResult | None:
        if not self._enabled:
            return None
        with self._lock:
            metrics = self._agents.get(agent_id)
            if metrics is None:
                return None
            self._stats["total_computes"] += 1
            components = {
                "task_completion": self._score_task_completion(metrics),
                "loop_freedom": self._score_loop_freedom(metrics),
                "cost_efficiency": self._score_cost_efficiency(metrics),
                "latency_health": self._score_latency_health(metrics),
                "context_stability": self._score_context_stability(metrics),
                "termination_quality": self._score_termination_quality(metrics),
                "security_compliance": self._score_security_compliance(metrics),
            }
            total_weight = sum(self._weights.values()) or 1
            score = sum(components[k] * self._weights.get(k, 0) / total_weight for k in components)
            score = round(min(100.0, max(0.0, score)), 2)
            if score >= 90:
                grade = "A"
            elif score >= 75:
                grade = "B"
            elif score >= 60:
                grade = "C"
            elif score >= 40:
                grade = "D"
            else:
                grade = "F"

            n = metrics.total_requests
            if n >= 200:
                confidence = "high"
            elif n >= 50:
                confidence = "medium"
            else:
                confidence = "low"

            return ARSResult(
                agent_id=agent_id,
                score=score,
                grade=grade,
                components={k: round(v, 2) for k, v in components.items()},
                sample_size=n,
                confidence=confidence,
                computed_at=time.time(),
            )

    def compute_all(self) -> list[ARSResult]:
        with self._lock:
            agent_ids = list(self._agents.keys())
        return [item for aid in agent_ids if (item := self.compute(aid)) is not None]

    def _score_task_completion(self, m: AgentMetrics) -> float:
        n = m.total_sessions
        if n == 0:
            return 50.0
        p = m.successful_sessions / n
        z = 1.96
        denominator = 1 + (z * z / n)
        center = p + (z * z / (2 * n))
        spread = z * math.sqrt((p * (1 - p) + z * z / (4 * n)) / n)
        lower = (center - spread) / denominator
        return round(min(100.0, max(0.0, lower * 100.0)), 2)

    def _score_loop_freedom(self, m: AgentMetrics) -> float:
        if m.total_requests == 0:
            return 50.0
        clean = m.total_requests - m.loop_flagged_requests
        rate = (clean + 1) / (m.total_requests + 2)
        return round(rate * 100.0, 2)

    def _score_cost_efficiency(self, m: AgentMetrics) -> float:
        if m.budget_ceiling_usd <= 0:
            return 50.0
        ratio = m.total_cost_usd / m.budget_ceiling_usd
        return round(max(0.0, 100.0 * (1.0 - min(1.0, ratio))), 2)

    def _score_latency_health(self, m: AgentMetrics) -> float:
        if len(m.latencies_ms) < 5:
            return 50.0
        sorted_lat = sorted(m.latencies_ms)
        idx = min(len(sorted_lat) - 1, int(len(sorted_lat) * 0.95))
        p95 = sorted_lat[idx]
        if p95 <= 0:
            return 100.0
        return round(max(0.0, 100.0 * min(1.0, self._latency_baseline / p95)), 2)

    def _score_context_stability(self, m: AgentMetrics) -> float:
        if len(m.token_counts) < 3:
            return 50.0
        n = len(m.token_counts)
        mean = sum(m.token_counts) / n
        if mean <= 0:
            return 50.0
        variance = sum((x - mean) ** 2 for x in m.token_counts) / n
        cv = math.sqrt(variance) / mean
        score = 100.0 * max(0.0, 1.0 - cv)
        return round(max(0.0, min(100.0, score)), 2)

    def _score_termination_quality(self, m: AgentMetrics) -> float:
        total = m.clean_terminations + m.dirty_terminations
        if total == 0:
            return 50.0
        return round((m.clean_terminations / total) * 100.0, 2)

    def _score_security_compliance(self, m: AgentMetrics) -> float:
        if m.total_requests == 0:
            return 50.0
        clean = m.total_requests - m.requests_with_threats
        return round((clean / m.total_requests) * 100.0, 2)

    def get_agent_metrics(self, agent_id: str) -> dict[str, Any] | None:
        with self._lock:
            metrics = self._agents.get(agent_id)
            if metrics is None:
                return None
            return {
                "agent_id": metrics.agent_id,
                "total_requests": metrics.total_requests,
                "total_sessions": metrics.total_sessions,
                "successful_sessions": metrics.successful_sessions,
                "total_cost_usd": round(metrics.total_cost_usd, 4),
                "loop_flagged_requests": metrics.loop_flagged_requests,
                "requests_with_threats": metrics.requests_with_threats,
                "clean_terminations": metrics.clean_terminations,
                "dirty_terminations": metrics.dirty_terminations,
                "first_seen": metrics.first_seen,
                "last_seen": metrics.last_seen,
            }

    @property
    def stats(self) -> dict[str, Any]:
        with self._lock:
            return {
                **self._stats,
                "enabled": self._enabled,
                "weights": dict(self._weights),
                "max_agents": self._max_agents,
                "active_agents": len(self._agents),
            }



"""OTLP HTTP/JSON span and metrics exporter using stdlib only."""

from __future__ import annotations

import json
import threading
import time
import urllib.error
import urllib.request
from collections import deque
from dataclasses import dataclass, field
from typing import Any

from orchesis.otel import SpanData


@dataclass
class OTLPExportConfig:
    """Configuration for OTLP export."""

    enabled: bool = False
    endpoint: str = "http://localhost:4318"
    traces_path: str = "/v1/traces"
    metrics_path: str = "/v1/metrics"
    headers: dict[str, str] = field(default_factory=dict)
    timeout_seconds: float = 10.0
    batch_size: int = 50
    flush_interval_seconds: float = 5.0
    max_queue_size: int = 2000
    retry_count: int = 2
    resource_attributes: dict[str, str] = field(
        default_factory=lambda: {
            "service.name": "orchesis-proxy",
            "service.version": "0.8.0",
        }
    )


class OTLPSpanExporter:
    """
    Batched OTLP HTTP/JSON span exporter.

    Sends spans to any OpenTelemetry-compatible backend:
    Grafana Tempo, Jaeger, Datadog, Honeycomb, etc.

    Uses stdlib urllib — no opentelemetry-sdk dependency.
    Thread-safe with background flush.
    """

    def __init__(self, config: OTLPExportConfig | None = None) -> None:
        self._config = config or OTLPExportConfig()
        self._lock = threading.Lock()
        self._queue: deque[SpanData] = deque(maxlen=self._config.max_queue_size)
        self._flush_thread: threading.Thread | None = None
        self._running = False
        self._exported: int = 0
        self._failed: int = 0
        self._dropped: int = 0

    def start(self) -> None:
        """Start background flush thread."""
        if self._running or not self._config.enabled:
            return
        self._running = True
        self._flush_thread = threading.Thread(
            target=self._flush_loop, daemon=True, name="orchesis-otlp-flush"
        )
        self._flush_thread.start()

    def stop(self) -> None:
        """Stop background flush and export remaining spans."""
        self._running = False
        if self._flush_thread and self._flush_thread.is_alive():
            self._flush_thread.join(timeout=self._config.timeout_seconds)
        self.flush()

    def export_span(self, span: SpanData) -> None:
        """Add span to export queue. Non-blocking."""
        with self._lock:
            if len(self._queue) >= self._config.max_queue_size:
                self._dropped += 1
            self._queue.append(span)
            if len(self._queue) >= self._config.batch_size:
                self._flush_batch()

    def flush(self) -> int:
        """Export all queued spans. Returns count exported."""
        with self._lock:
            return self._flush_batch()

    def _flush_loop(self) -> None:
        """Background thread: flush periodically."""
        while self._running:
            time.sleep(self._config.flush_interval_seconds)
            with self._lock:
                if self._queue:
                    self._flush_batch()

    def _flush_batch(self) -> int:
        """Export current queue as OTLP batch. Must hold lock."""
        if not self._queue:
            return 0
        batch = list(self._queue)
        self._queue.clear()
        payload = self._build_otlp_payload(batch)
        success = self._send_http(payload)
        if success:
            self._exported += len(batch)
        else:
            self._failed += len(batch)
        return len(batch)

    def _build_otlp_payload(self, spans: list[SpanData]) -> dict[str, Any]:
        """
        Build OTLP JSON payload following the spec:
        https://opentelemetry.io/docs/specs/otlp/#otlphttp
        """
        resource_attrs = [
            {"key": k, "value": {"stringValue": str(v)}}
            for k, v in self._config.resource_attributes.items()
        ]
        otlp_spans = []
        for span in spans:
            otlp_attrs = []
            for key, value in span.attributes.items():
                if isinstance(value, bool):
                    otlp_attrs.append({"key": key, "value": {"boolValue": value}})
                elif isinstance(value, int):
                    otlp_attrs.append({"key": key, "value": {"intValue": str(value)}})
                elif isinstance(value, float):
                    otlp_attrs.append({"key": key, "value": {"doubleValue": value}})
                else:
                    otlp_attrs.append({"key": key, "value": {"stringValue": str(value)}})
            otlp_events = []
            for evt in span.events:
                evt_attrs = []
                for ek, ev in (evt.get("attributes") or {}).items():
                    if isinstance(ev, bool):
                        evt_attrs.append({"key": ek, "value": {"boolValue": ev}})
                    elif isinstance(ev, int):
                        evt_attrs.append({"key": ek, "value": {"intValue": str(ev)}})
                    else:
                        evt_attrs.append({"key": ek, "value": {"stringValue": str(ev)}})
                otlp_events.append({
                    "name": evt.get("name", ""),
                    "timeUnixNano": str(evt.get("timestamp_ns", 0)),
                    "attributes": evt_attrs,
                })
            status_code = 1 if span.status == "OK" else 2
            trace_id = span.trace_id.ljust(32, "0")[:32]
            span_id = span.span_id.ljust(16, "0")[:16]
            parent_span_id = (
                (span.parent_span_id or "").ljust(16, "0")[:16]
                if span.parent_span_id
                else ""
            )
            otlp_spans.append({
                "traceId": trace_id,
                "spanId": span_id,
                "parentSpanId": parent_span_id,
                "name": span.operation,
                "kind": 2,
                "startTimeUnixNano": str(span.start_time_ns),
                "endTimeUnixNano": str(span.end_time_ns),
                "attributes": otlp_attrs,
                "events": otlp_events,
                "status": {"code": status_code, "message": span.status},
            })
        return {
            "resourceSpans": [{
                "resource": {"attributes": resource_attrs},
                "scopeSpans": [{
                    "scope": {
                        "name": "orchesis",
                        "version": self._config.resource_attributes.get("service.version", "0.8.0"),
                    },
                    "spans": otlp_spans,
                }],
            }]
        }

    def _send_http(self, payload: dict[str, Any]) -> bool:
        """Send OTLP payload via HTTP POST. Returns success."""
        url = f"{self._config.endpoint.rstrip('/')}{self._config.traces_path}"
        body = json.dumps(payload, ensure_ascii=False).encode("utf-8")
        headers = {
            "Content-Type": "application/json",
            **self._config.headers,
        }
        for attempt in range(max(1, self._config.retry_count)):
            try:
                req = urllib.request.Request(url, data=body, headers=headers, method="POST")
                with urllib.request.urlopen(req, timeout=self._config.timeout_seconds) as resp:
                    if resp.status < 300:
                        return True
            except (urllib.error.URLError, urllib.error.HTTPError, OSError, TimeoutError):
                if attempt < self._config.retry_count - 1:
                    time.sleep(0.5 * (attempt + 1))
                continue
        return False

    def get_stats(self) -> dict[str, Any]:
        """Return exporter statistics."""
        with self._lock:
            return {
                "enabled": self._config.enabled,
                "endpoint": self._config.endpoint,
                "queue_size": len(self._queue),
                "exported": self._exported,
                "failed": self._failed,
                "dropped": self._dropped,
            }



"""Agent discovery and inventory for Orchesis."""

from __future__ import annotations

from dataclasses import asdict, dataclass, field
import threading
import time
from typing import Any, Optional


@dataclass
class AgentProfile:
    """Complete profile of a discovered agent."""

    agent_id: str
    first_seen: float
    last_seen: float
    total_requests: int
    total_tokens: int
    total_cost_usd: float
    models_used: list[str]
    primary_model: str
    tools_used: list[str]
    tool_call_count: int
    ars_grade: str
    ars_score: float
    anomaly_count: int
    mast_findings_count: int
    risk_level: str
    avg_tokens_per_request: float
    avg_latency_ms: float
    avg_cost_per_request: float
    session_count: int
    is_cron: bool
    status: str
    health: str


class AgentDiscovery:
    """Centralized agent inventory and status."""

    def __init__(self, config: Optional[dict] = None):
        cfg = config if isinstance(config, dict) else {}
        self.enabled = bool(cfg.get("enabled", True))
        self.retention_hours = max(1, int(cfg.get("retention_hours", 168)))
        thresholds = cfg.get("health_thresholds") if isinstance(cfg.get("health_thresholds"), dict) else {}
        self.warning_anomaly_rate = float(thresholds.get("warning_anomaly_rate", 0.1))
        self.critical_anomaly_rate = float(thresholds.get("critical_anomaly_rate", 0.3))
        self._lock = threading.Lock()
        self._agents: dict[str, dict[str, Any]] = {}
        self._stats = {
            "requests_recorded": 0,
            "detections_recorded": 0,
            "cleanup_runs": 0,
            "agents_removed": 0,
        }

    @staticmethod
    def _now() -> float:
        return time.time()

    def _get_or_create(self, agent_id: str) -> dict[str, Any]:
        key = str(agent_id or "unknown")
        now = self._now()
        item = self._agents.get(key)
        if item is None:
            item = {
                "agent_id": key,
                "first_seen": now,
                "last_seen": now,
                "total_requests": 0,
                "total_tokens": 0,
                "total_cost_usd": 0.0,
                "total_latency_ms": 0.0,
                "models_counter": {},
                "tools_counter": {},
                "tool_call_count": 0,
                "ars_grade": "",
                "ars_score": 0.0,
                "anomaly_count": 0,
                "mast_findings_count": 0,
                "risk_level": "low",
                "sessions": set(),
                "is_cron": False,
                "blocked": False,
                "rate_limited": False,
            }
            self._agents[key] = item
        return item

    def record_request(
        self,
        agent_id: str,
        request_data: dict,
        model: str | None = None,
        tokens: int = 0,
        cost: float = 0.0,
        latency_ms: float = 0.0,
        tools: list | None = None,
    ) -> None:
        if not self.enabled:
            return
        req = request_data if isinstance(request_data, dict) else {}
        with self._lock:
            item = self._get_or_create(agent_id)
            now = self._now()
            item["last_seen"] = now
            item["total_requests"] += 1
            item["total_tokens"] += max(0, int(tokens or 0))
            item["total_cost_usd"] += max(0.0, float(cost or 0.0))
            item["total_latency_ms"] += max(0.0, float(latency_ms or 0.0))
            model_name = str(model or req.get("model", "") or "")
            if model_name:
                models_counter = item["models_counter"]
                models_counter[model_name] = int(models_counter.get(model_name, 0)) + 1
            session_id = req.get("session_id")
            if isinstance(session_id, str) and session_id:
                item["sessions"].add(session_id)
            parsed_tools: list[str] = []
            if isinstance(tools, list):
                for tool in tools:
                    if isinstance(tool, str) and tool:
                        parsed_tools.append(tool)
                    elif isinstance(tool, dict):
                        name = tool.get("name") or (tool.get("function") or {}).get("name")
                        if isinstance(name, str) and name:
                            parsed_tools.append(name)
            for name in parsed_tools:
                tools_counter = item["tools_counter"]
                tools_counter[name] = int(tools_counter.get(name, 0)) + 1
                item["tool_call_count"] += 1
            self._stats["requests_recorded"] += 1

    def record_detection(
        self,
        agent_id: str,
        anomaly_score: float = 0.0,
        mast_findings: int = 0,
        risk_level: str = "low",
        ars_grade: str | None = None,
        ars_score: float | None = None,
        is_cron: bool = False,
        status: str | None = None,
    ) -> None:
        if not self.enabled:
            return
        with self._lock:
            item = self._get_or_create(agent_id)
            item["last_seen"] = self._now()
            if float(anomaly_score or 0.0) > 0.0:
                item["anomaly_count"] += 1
            item["mast_findings_count"] += max(0, int(mast_findings or 0))
            safe_risk = str(risk_level or "low").lower()
            if safe_risk in {"low", "medium", "high", "critical"}:
                item["risk_level"] = safe_risk
            if isinstance(ars_grade, str):
                item["ars_grade"] = ars_grade
            if isinstance(ars_score, int | float):
                item["ars_score"] = float(ars_score)
            if is_cron:
                item["is_cron"] = True
            if isinstance(status, str) and status in {"active", "idle", "rate_limited", "blocked"}:
                item["blocked"] = status == "blocked"
                item["rate_limited"] = status == "rate_limited"
            elif safe_risk == "critical":
                item["blocked"] = True
            self._stats["detections_recorded"] += 1

    def _derive_status(self, item: dict[str, Any], now: float) -> str:
        if bool(item.get("blocked", False)):
            return "blocked"
        if bool(item.get("rate_limited", False)):
            return "rate_limited"
        idle_for = now - float(item.get("last_seen", now))
        if idle_for > 3600.0:
            return "idle"
        return "active"

    def _derive_health(self, item: dict[str, Any]) -> str:
        total_requests = max(1, int(item.get("total_requests", 0)))
        anomaly_rate = float(item.get("anomaly_count", 0)) / float(total_requests)
        risk_level = str(item.get("risk_level", "low")).lower()
        if risk_level == "critical" or anomaly_rate >= self.critical_anomaly_rate:
            return "critical"
        if risk_level == "high" or anomaly_rate >= self.warning_anomaly_rate:
            return "warning"
        return "healthy"

    @staticmethod
    def _top_key(counter: dict[str, int]) -> str:
        if not counter:
            return ""
        return max(counter.items(), key=lambda item: item[1])[0]

    def _to_profile(self, item: dict[str, Any]) -> AgentProfile:
        now = self._now()
        total_requests = int(item.get("total_requests", 0))
        total_tokens = int(item.get("total_tokens", 0))
        total_cost = float(item.get("total_cost_usd", 0.0))
        total_latency = float(item.get("total_latency_ms", 0.0))
        models_counter = dict(item.get("models_counter", {}))
        tools_counter = dict(item.get("tools_counter", {}))
        status = self._derive_status(item, now)
        health = self._derive_health(item)
        denom = float(total_requests) if total_requests > 0 else 1.0
        return AgentProfile(
            agent_id=str(item.get("agent_id", "unknown")),
            first_seen=float(item.get("first_seen", now)),
            last_seen=float(item.get("last_seen", now)),
            total_requests=total_requests,
            total_tokens=total_tokens,
            total_cost_usd=round(total_cost, 8),
            models_used=sorted(models_counter.keys()),
            primary_model=self._top_key(models_counter),
            tools_used=sorted(tools_counter.keys()),
            tool_call_count=int(item.get("tool_call_count", 0)),
            ars_grade=str(item.get("ars_grade", "") or ""),
            ars_score=float(item.get("ars_score", 0.0) or 0.0),
            anomaly_count=int(item.get("anomaly_count", 0)),
            mast_findings_count=int(item.get("mast_findings_count", 0)),
            risk_level=str(item.get("risk_level", "low") or "low"),
            avg_tokens_per_request=round(total_tokens / denom, 3),
            avg_latency_ms=round(total_latency / denom, 3),
            avg_cost_per_request=round(total_cost / denom, 8),
            session_count=len(item.get("sessions", set())),
            is_cron=bool(item.get("is_cron", False)),
            status=status,
            health=health,
        )

    def get_agent(self, agent_id: str) -> Optional[AgentProfile]:
        with self._lock:
            raw = self._agents.get(str(agent_id or "unknown"))
            if raw is None:
                return None
            snapshot = dict(raw)
            snapshot["models_counter"] = dict(raw.get("models_counter", {}))
            snapshot["tools_counter"] = dict(raw.get("tools_counter", {}))
            snapshot["sessions"] = set(raw.get("sessions", set()))
        return self._to_profile(snapshot)

    def get_all_agents(self) -> list[AgentProfile]:
        with self._lock:
            snapshots: list[dict[str, Any]] = []
            for raw in self._agents.values():
                snap = dict(raw)
                snap["models_counter"] = dict(raw.get("models_counter", {}))
                snap["tools_counter"] = dict(raw.get("tools_counter", {}))
                snap["sessions"] = set(raw.get("sessions", set()))
                snapshots.append(snap)
        profiles = [self._to_profile(item) for item in snapshots]
        profiles.sort(key=lambda item: item.last_seen, reverse=True)
        return profiles

    def get_summary(self) -> dict:
        profiles = self.get_all_agents()
        now = self._now()
        active_cutoff = now - 3600.0
        health = {"healthy": 0, "warning": 0, "critical": 0}
        models: set[str] = set()
        total_cost_24h = 0.0
        total_requests_24h = 0
        for item in profiles:
            health[item.health] = int(health.get(item.health, 0)) + 1
            for model in item.models_used:
                models.add(model)
            if item.last_seen >= (now - 86400.0):
                total_cost_24h += float(item.total_cost_usd)
                total_requests_24h += int(item.total_requests)
        top_by_cost = sorted(profiles, key=lambda item: item.total_cost_usd, reverse=True)[:5]
        top_by_risk = sorted(
            profiles,
            key=lambda item: (float(item.anomaly_count) / float(max(1, item.total_requests))),
            reverse=True,
        )[:5]
        return {
            "total_agents": len(profiles),
            "active_agents": sum(1 for item in profiles if item.last_seen >= active_cutoff and item.status == "active"),
            "idle_agents": sum(1 for item in profiles if item.status == "idle"),
            "blocked_agents": sum(1 for item in profiles if item.status == "blocked"),
            "health": health,
            "total_cost_24h_usd": round(total_cost_24h, 6),
            "total_requests_24h": int(total_requests_24h),
            "models_in_use": sorted(models),
            "top_agents_by_cost": [asdict(item) for item in top_by_cost],
            "top_agents_by_risk": [asdict(item) for item in top_by_risk],
        }

    def cleanup(self) -> int:
        cutoff = self._now() - (self.retention_hours * 3600.0)
        removed = 0
        with self._lock:
            for key in list(self._agents.keys()):
                if float(self._agents[key].get("last_seen", 0.0)) < cutoff:
                    self._agents.pop(key, None)
                    removed += 1
            self._stats["cleanup_runs"] += 1
            self._stats["agents_removed"] += removed
        return removed

    def get_stats(self) -> dict:
        with self._lock:
            return {
                "enabled": self.enabled,
                "agents_tracked": len(self._agents),
                **self._stats,
            }

    def reset(self) -> None:
        with self._lock:
            self._agents = {}
            self._stats = {
                "requests_recorded": 0,
                "detections_recorded": 0,
                "cleanup_runs": 0,
                "agents_removed": 0,
            }




from __future__ import annotations

import os
import shutil
import shlex
import sys
from contextlib import asynccontextmanager
from datetime import timedelta
from pathlib import Path

import pytest
from mcp.client.session import ClientSession
from mcp.client.stdio import StdioServerParameters, stdio_client

pytestmark = pytest.mark.skipif(
    shutil.which("orchesis") is None,
    reason="requires orchesis CLI installed",
)


def _sqlite_server_command() -> tuple[str, list[str]]:
    python_dir = Path(sys.executable).resolve().parent
    exe = python_dir / "Scripts" / "mcp-server-sqlite.exe"
    if not exe.exists():
        exe = python_dir.with_name("Scripts") / "mcp-server-sqlite.exe"
    if not exe.exists():
        pytest.skip("mcp-server-sqlite executable is not available")
    return str(exe), []


def _write_policy(path: Path) -> None:
    path.write_text(
        """
rules:
  - name: sql_restriction
    denied_operations:
      - DROP
      - DELETE
      - TRUNCATE
      - ALTER
  - name: file_access
    allowed_paths:
      - "/data"
      - "/tmp"
    denied_paths:
      - "/etc"
      - "/root"
""".strip(),
        encoding="utf-8",
    )


@asynccontextmanager
async def _proxy_session(
    *,
    policy_path: Path,
    downstream_command: str,
    downstream_args: list[str],
    timeout_seconds: float | None = None,
):
    env = os.environ.copy()
    env["POLICY_PATH"] = str(policy_path)
    env["DOWNSTREAM_COMMAND"] = downstream_command
    env["DOWNSTREAM_ARGS"] = " ".join(shlex.quote(arg) for arg in downstream_args)
    if timeout_seconds is not None:
        env["DOWNSTREAM_TIMEOUT_SECONDS"] = str(timeout_seconds)

    params = StdioServerParameters(
        command=sys.executable,
        args=["-m", "orchesis.mcp_proxy"],
        env=env,
        cwd=Path(__file__).resolve().parents[1],
    )
    async with stdio_client(params) as (read_stream, write_stream):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()
            yield session


@pytest.mark.asyncio
async def test_real_sqlite_tools_passthrough(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    db_path = tmp_path / "demo.db"
    _write_policy(policy_path)
    command, base_args = _sqlite_server_command()

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=command,
        downstream_args=[*base_args, "--db-path", str(db_path)],
    ) as session:
        tools = await session.list_tools()
        names = {tool.name for tool in tools.tools}

    assert "read_query" in names
    assert "write_query" in names


@pytest.mark.asyncio
async def test_real_sqlite_allow_safe_read_operation(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    db_path = tmp_path / "demo.db"
    _write_policy(policy_path)
    command, base_args = _sqlite_server_command()

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=command,
        downstream_args=[*base_args, "--db-path", str(db_path)],
    ) as session:
        result = await session.call_tool("read_query", {"query": "SELECT 1 AS n"})

    assert result.isError is False


@pytest.mark.asyncio
async def test_real_sqlite_deny_delete_operation(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    db_path = tmp_path / "demo.db"
    _write_policy(policy_path)
    command, base_args = _sqlite_server_command()

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=command,
        downstream_args=[*base_args, "--db-path", str(db_path)],
    ) as session:
        result = await session.call_tool("write_query", {"query": "DELETE FROM users"})

    assert result.isError is True
    assert "DELETE is denied" in result.content[0].text


@pytest.mark.asyncio
async def test_real_sqlite_malformed_arguments_are_graceful(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    db_path = tmp_path / "demo.db"
    _write_policy(policy_path)
    command, base_args = _sqlite_server_command()

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=command,
        downstream_args=[*base_args, "--db-path", str(db_path)],
    ) as session:
        result = await session.call_tool("read_query", {"query": 123})

    assert result.isError is True


@pytest.mark.asyncio
async def test_proxy_handles_downstream_timeout(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    slow_server = tmp_path / "slow_server.py"
    _write_policy(policy_path)
    slow_server.write_text(
        """
import asyncio
import mcp.server.stdio
import mcp.types as types
from mcp.server.lowlevel import NotificationOptions, Server
from mcp.server.models import InitializationOptions

server = Server("slow-server")

@server.list_tools()
async def list_tools():
    return [types.Tool(name="slow_tool", inputSchema={"type":"object","properties":{}})]

@server.call_tool()
async def call_tool(name, arguments):
    _ = (name, arguments)
    await asyncio.sleep(5)
    return [types.TextContent(type="text", text="done")]

async def run():
    async with mcp.server.stdio.stdio_server() as (r, w):
        await server.run(
            r,
            w,
            InitializationOptions(
                server_name="slow-server",
                server_version="0.1.0",
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(),
                    experimental_capabilities={}
                ),
            ),
        )

asyncio.run(run())
""".strip(),
        encoding="utf-8",
    )

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=sys.executable,
        downstream_args=[str(slow_server)],
        timeout_seconds=0.5,
    ) as session:
        result = await session.call_tool(
            "slow_tool",
            {},
            read_timeout_seconds=timedelta(seconds=3),
        )

    assert result.isError is True
    assert "timeout" in result.content[0].text.lower()


@pytest.mark.asyncio
async def test_proxy_handles_downstream_crash(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    crash_server = tmp_path / "crash_server.py"
    _write_policy(policy_path)
    crash_server.write_text(
        """
import os
import mcp.server.stdio
import mcp.types as types
from mcp.server.lowlevel import NotificationOptions, Server
from mcp.server.models import InitializationOptions

server = Server("crash-server")

@server.list_tools()
async def list_tools():
    return [types.Tool(name="crash_tool", inputSchema={"type":"object","properties":{}})]

@server.call_tool()
async def call_tool(name, arguments):
    _ = (name, arguments)
    os._exit(1)

async def run():
    async with mcp.server.stdio.stdio_server() as (r, w):
        await server.run(
            r,
            w,
            InitializationOptions(
                server_name="crash-server",
                server_version="0.1.0",
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(),
                    experimental_capabilities={}
                ),
            ),
        )

import asyncio
asyncio.run(run())
""".strip(),
        encoding="utf-8",
    )

    async with _proxy_session(
        policy_path=policy_path,
        downstream_command=sys.executable,
        downstream_args=[str(crash_server)],
    ) as session:
        result = await session.call_tool(
            "crash_tool",
            {},
            read_timeout_seconds=timedelta(seconds=3),
        )

    assert result.isError is True



"""Community Intelligence Network client."""

from __future__ import annotations

from dataclasses import asdict, dataclass
import json
import logging
from pathlib import Path
import sqlite3
import threading
import time
import uuid
from typing import Any, Optional
from urllib.error import HTTPError, URLError
from urllib.request import Request as UrlRequest, urlopen

from orchesis.privacy_filter import (
    CommunitySignal,
    PRIVACY_LEVEL_STANDARD,
    PrivacyFilter,
)

_COMMUNITY_LOGGER = logging.getLogger("orchesis.community")


@dataclass
class CommunitySignature:
    """A threat signature received from the community."""

    signature_id: str
    signature_type: str
    pattern: str
    confidence: float
    reporters: int
    first_seen: float
    last_seen: float
    severity: str


@dataclass
class CommunityStats:
    """Statistics about community participation."""

    trust_score: float
    signals_sent: int
    signals_pending: int
    signatures_received: int
    signatures_cached: int
    last_sync: float
    hub_status: str
    contribution_ratio: float
    community_size: int


class CommunityClient:
    """Client for the Orchesis Community Intelligence Network."""

    def __init__(self, config: Optional[dict] = None):
        cfg = config if isinstance(config, dict) else {}
        self.enabled = bool(cfg.get("enabled", True))
        self.privacy_level = int(cfg.get("privacy_level", PRIVACY_LEVEL_STANDARD))
        self.hub_url = str(cfg.get("hub_url", "https://community.orchesis.io/api/v1")).rstrip("/")
        self.send_interval_seconds = max(0.1, float(cfg.get("send_interval_seconds", 300.0)))
        self.poll_interval_seconds = max(0.1, float(cfg.get("poll_interval_seconds", 900.0)))
        self.max_batch_size = max(1, int(cfg.get("max_batch_size", 100)))
        self.max_pending_signals = max(1, int(cfg.get("max_pending_signals", 10000)))
        self.max_cached_signatures = max(100, int(cfg.get("max_cached_signatures", 5000)))
        self.min_anomaly_score = max(0.0, float(cfg.get("min_anomaly_score", 30.0)))
        self._hub_status = "offline"
        self._trust_score = 0.0
        self._community_size = 0
        self._signals_sent = 0
        self._signatures_received = 0
        self._last_sync = 0.0
        self._lock = threading.RLock()
        self._stop_event = threading.Event()
        self._sender_thread: threading.Thread | None = None
        self._receiver_thread: threading.Thread | None = None
        self._privacy_filter = PrivacyFilter(self.privacy_level, min_anomaly_score=self.min_anomaly_score)
        self._data_dir = Path(str(cfg.get("data_dir", ".orchesis/community")))
        self._instance_file = self._data_dir / "instance_id"
        self._db_path = self._data_dir / "community.sqlite3"
        self.instance_id = ""
        if not self.enabled:
            return
        self._data_dir.mkdir(parents=True, exist_ok=True)
        self.instance_id = self._get_or_create_instance_id()
        self._init_db()

    def _connect(self) -> sqlite3.Connection:
        conn = sqlite3.connect(str(self._db_path), timeout=30.0, check_same_thread=False)
        conn.row_factory = sqlite3.Row
        return conn

    def _set_metadata(self, key: str, value: str) -> None:
        with self._connect() as conn:
            conn.execute(
                "INSERT INTO metadata(key, value) VALUES(?, ?) ON CONFLICT(key) DO UPDATE SET value=excluded.value",
                (key, value),
            )
            conn.commit()

    def _get_metadata(self, key: str, default: str = "") -> str:
        with self._connect() as conn:
            row = conn.execute("SELECT value FROM metadata WHERE key = ?", (key,)).fetchone()
            return str(row["value"]) if row else default

    def record_signal(self, signal: CommunitySignal) -> None:
        if not self.enabled:
            return
        try:
            if not self._privacy_filter.validate_signal(signal):
                return
            payload = json.dumps(asdict(signal), ensure_ascii=False)
            now = time.time()
            with self._lock:
                with self._connect() as conn:
                    conn.execute(
                        "INSERT INTO pending_signals(signal_json, created_at, retry_count, sent_at) VALUES(?, ?, 0, NULL)",
                        (payload, now),
                    )
                    count_row = conn.execute("SELECT COUNT(*) AS c FROM pending_signals WHERE sent_at IS NULL").fetchone()
                    pending = int(count_row["c"]) if count_row else 0
                    if pending > self.max_pending_signals:
                        to_drop = pending - self.max_pending_signals
                        conn.execute(
                            "DELETE FROM pending_signals WHERE id IN (SELECT id FROM pending_signals WHERE sent_at IS NULL ORDER BY created_at ASC LIMIT ?)",
                            (to_drop,),
                        )
                    conn.commit()
        except Exception as exc:  # noqa: BLE001
            _COMMUNITY_LOGGER.warning("record_signal failed: %s", exc)

    def record_detection(
        self,
        detection_result: Any,
        telemetry_record: Any = None,
        ars_data: Any = None,
        request_meta: Any = None,
    ) -> None:
        if not self.enabled:
            return
        try:
            signal = self._privacy_filter.create_signal(
                detection_result=detection_result,
                telemetry_record=telemetry_record,
                ars_data=ars_data,
                request_meta=request_meta,
            )
            if signal is None:
                return
            if float(signal.anomaly_score) < self.min_anomaly_score and not signal.threat_ids:
                return
            self.record_signal(signal)
        except Exception as exc:  # noqa: BLE001
            _COMMUNITY_LOGGER.warning("record_detection failed: %s", exc)

    def _post_json(self, url: str, payload: dict[str, Any]) -> tuple[int, dict[str, Any], dict[str, str]]:
        data = json.dumps(payload, ensure_ascii=False).encode("utf-8")
        req = UrlRequest(
            url,
            data=data,
            headers={"Content-Type": "application/json", "X-Orchesis-Instance": self.instance_id},
            method="POST",
        )
        with urlopen(req, timeout=5) as resp:
            raw = resp.read().decode("utf-8") if resp.length != 0 else "{}"
            body = json.loads(raw) if raw.strip() else {}
            return int(resp.status), body if isinstance(body, dict) else {}, dict(resp.headers.items())

    def _get_json(self, url: str) -> tuple[int, Any]:
        req = UrlRequest(url, headers={"X-Orchesis-Instance": self.instance_id}, method="GET")
        with urlopen(req, timeout=5) as resp:
            raw = resp.read().decode("utf-8")
            payload = json.loads(raw) if raw.strip() else {}
            return int(resp.status), payload

    def _send_batch(self) -> tuple[int, int]:
        if not self.enabled:
            return (0, 0)
        try:
            with self._lock:
                with self._connect() as conn:
                    rows = conn.execute(
                        "SELECT id, signal_json FROM pending_signals WHERE sent_at IS NULL ORDER BY created_at ASC LIMIT ?",
                        (self.max_batch_size,),
                    ).fetchall()
            if not rows:
                return (0, 0)
            signals: list[dict[str, Any]] = []
            ids: list[int] = []
            for row in rows:
                ids.append(int(row["id"]))
                try:
                    parsed = json.loads(str(row["signal_json"]))
                    if isinstance(parsed, dict):
                        signals.append(parsed)
                except Exception:
                    continue
            if not signals:
                return (0, 0)
            status, body, _headers = self._post_json(
                f"{self.hub_url}/signals",
                {"instance_id": self.instance_id, "signals": signals},
            )
            if 200 <= status < 300:
                now = time.time()
                with self._lock:
                    with self._connect() as conn:
                        conn.executemany("UPDATE pending_signals SET sent_at = ? WHERE id = ?", [(now, sid) for sid in ids])
                        conn.commit()
                sent = len(ids)
                self._signals_sent += sent
                self._hub_status = "connected"
                self._trust_score = float(body.get("trust_score", self._trust_score))
                self._community_size = int(body.get("community_size", self._community_size))
                return (sent, 0)
            with self._lock:
                with self._connect() as conn:
                    conn.executemany(
                        "UPDATE pending_signals SET retry_count = retry_count + 1 WHERE id = ?",
                        [(sid,) for sid in ids],
                    )
                    conn.commit()
            self._hub_status = "error"
            return (0, len(ids))
        except HTTPError as exc:
            self._hub_status = "error"
            if int(exc.code) == 429:
                retry_after = 0.0
                try:
                    retry_after = float(exc.headers.get("Retry-After", "0") or "0")
                except Exception:
                    retry_after = 0.0
                if retry_after > 0:
                    time.sleep(min(1.0, retry_after))
                return (0, 0)
            _COMMUNITY_LOGGER.warning("community send failed: %s", exc)
            return (0, 0)
        except (URLError, OSError) as exc:
            self._hub_status = "offline"
            _COMMUNITY_LOGGER.warning("community hub unreachable: %s", exc)
            return (0, 0)
        except Exception as exc:  # noqa: BLE001
            self._hub_status = "error"
            _COMMUNITY_LOGGER.warning("send_batch error: %s", exc)
            return (0, 0)

    def _poll_signatures(self) -> int:
        if not self.enabled:
            return 0
        try:
            since = int(self._last_sync or 0)
            status, payload = self._get_json(f"{self.hub_url}/signatures?since={since}")
            if status < 200 or status >= 300:
                self._hub_status = "error"
                return 0
            items = payload.get("signatures", payload) if isinstance(payload, dict) else payload
            if not isinstance(items, list):
                return 0
            accepted = 0
            now = time.time()
            with self._lock:
                with self._connect() as conn:
                    for raw in items:
                        if not isinstance(raw, dict):
                            continue
                        sid = str(raw.get("signature_id", "") or "")
                        if not sid:
                            continue
                        sig = {
                            "signature_id": sid,
                            "signature_type": str(raw.get("signature_type", "") or ""),
                            "pattern": str(raw.get("pattern", "") or ""),
                            "confidence": float(raw.get("confidence", 0.0) or 0.0),
                            "reporters": int(raw.get("reporters", 0) or 0),
                            "first_seen": float(raw.get("first_seen", now) or now),
                            "last_seen": float(raw.get("last_seen", now) or now),
                            "severity": str(raw.get("severity", "low") or "low"),
                        }
                        expires = raw.get("expires_at")
                        expires_at = float(expires) if isinstance(expires, (int, float)) else None
                        conn.execute(
                            "INSERT INTO community_signatures(signature_id, signature_json, received_at, expires_at) VALUES(?, ?, ?, ?) "
                            "ON CONFLICT(signature_id) DO UPDATE SET signature_json=excluded.signature_json, received_at=excluded.received_at, expires_at=excluded.expires_at",
                            (sid, json.dumps(sig, ensure_ascii=False), now, expires_at),
                        )
                        accepted += 1
                    row = conn.execute("SELECT COUNT(*) AS c FROM community_signatures").fetchone()
                    total_cached = int(row["c"]) if row else 0
                    if total_cached > self.max_cached_signatures:
                        drop = total_cached - self.max_cached_signatures
                        conn.execute(
                            "DELETE FROM community_signatures WHERE signature_id IN (SELECT signature_id FROM community_signatures ORDER BY received_at ASC LIMIT ?)",
                            (drop,),
                        )
                    conn.commit()
            self._last_sync = now
            self._set_metadata("last_sync", str(now))
            self._signatures_received += accepted
            self._hub_status = "connected"
            return accepted
        except (URLError, OSError) as exc:
            self._hub_status = "offline"
            _COMMUNITY_LOGGER.warning("community poll offline: %s", exc)
            return 0
        except Exception as exc:  # noqa: BLE001
            self._hub_status = "error"
            _COMMUNITY_LOGGER.warning("poll_signatures error: %s", exc)
            return 0

    def get_community_signatures(self) -> list[CommunitySignature]:
        if not self.enabled:
            return []
        now = time.time()
        out: list[CommunitySignature] = []
        with self._lock:
            with self._connect() as conn:
                rows = conn.execute(
                    "SELECT signature_json, expires_at FROM community_signatures ORDER BY received_at DESC"
                ).fetchall()
        for row in rows:
            expires_at = row["expires_at"]
            if isinstance(expires_at, (int, float)) and float(expires_at) > 0.0 and float(expires_at) < now:
                continue
            try:
                parsed = json.loads(str(row["signature_json"]))
                if not isinstance(parsed, dict):
                    continue
                out.append(
                    CommunitySignature(
                        signature_id=str(parsed.get("signature_id", "")),
                        signature_type=str(parsed.get("signature_type", "")),
                        pattern=str(parsed.get("pattern", "")),
                        confidence=float(parsed.get("confidence", 0.0) or 0.0),
                        reporters=int(parsed.get("reporters", 0) or 0),
                        first_seen=float(parsed.get("first_seen", 0.0) or 0.0),
                        last_seen=float(parsed.get("last_seen", 0.0) or 0.0),
                        severity=str(parsed.get("severity", "low") or "low"),
                    )
                )
            except Exception:
                continue
        return out

    def has_signature(self, signature_id: str) -> bool:
        if not self.enabled:
            return False
        with self._lock:
            with self._connect() as conn:
                row = conn.execute(
                    "SELECT 1 AS ok FROM community_signatures WHERE signature_id = ? LIMIT 1",
                    (str(signature_id),),
                ).fetchone()
                return row is not None

    def _sender_loop(self) -> None:
        while not self._stop_event.wait(self.send_interval_seconds):
            try:
                self._send_batch()
                self._cleanup_old_signals()
            except Exception as exc:  # noqa: BLE001
                _COMMUNITY_LOGGER.warning("community sender loop error: %s", exc)

    def _receiver_loop(self) -> None:
        while not self._stop_event.wait(self.poll_interval_seconds):
            try:
                self._poll_signatures()
                self._cleanup_old_signals()
            except Exception as exc:  # noqa: BLE001
                _COMMUNITY_LOGGER.warning("community receiver loop error: %s", exc)

    def start(self) -> None:
        if not self.enabled:
            return
        if self._sender_thread is not None and self._sender_thread.is_alive():
            return
        self._stop_event.clear()
        self._sender_thread = threading.Thread(target=self._sender_loop, name="community-sender", daemon=True)
        self._receiver_thread = threading.Thread(target=self._receiver_loop, name="community-receiver", daemon=True)
        self._sender_thread.start()
        self._receiver_thread.start()

    def stop(self) -> None:
        if not self.enabled:
            return
        self._stop_event.set()
        if self._sender_thread is not None:
            self._sender_thread.join(timeout=3.0)
        if self._receiver_thread is not None:
            self._receiver_thread.join(timeout=3.0)
        try:
            self._send_batch()
        except Exception:
            pass
        self._cleanup_old_signals()

    def _get_or_create_instance_id(self) -> str:
        if self._instance_file.exists():
            existing = self._instance_file.read_text(encoding="utf-8").strip()
            if existing:
                return existing
        value = str(uuid.uuid4())
        self._instance_file.write_text(value, encoding="utf-8")
        return value

    def _init_db(self) -> None:
        try:
            with self._connect() as conn:
                conn.execute("PRAGMA journal_mode=WAL")
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS pending_signals ("
                    "id INTEGER PRIMARY KEY,"
                    "signal_json TEXT NOT NULL,"
                    "created_at REAL NOT NULL,"
                    "retry_count INTEGER DEFAULT 0,"
                    "sent_at REAL DEFAULT NULL)"
                )
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS community_signatures ("
                    "signature_id TEXT PRIMARY KEY,"
                    "signature_json TEXT NOT NULL,"
                    "received_at REAL NOT NULL,"
                    "expires_at REAL DEFAULT NULL)"
                )
                conn.execute(
                    "CREATE TABLE IF NOT EXISTS metadata ("
                    "key TEXT PRIMARY KEY,"
                    "value TEXT NOT NULL)"
                )
                conn.commit()
            last_sync = self._get_metadata("last_sync", "0")
            try:
                self._last_sync = float(last_sync)
            except Exception:
                self._last_sync = 0.0
        except sqlite3.DatabaseError:
            recovered_path = self._data_dir / f"community-recovered-{int(time.time() * 1000)}.sqlite3"
            try:
                if self._db_path.exists():
                    backup = self._db_path.with_suffix(".corrupt")
                    try:
                        self._db_path.replace(backup)
                    except Exception:
                        self._db_path.unlink(missing_ok=True)
            except Exception:
                pass
            self._db_path = recovered_path
            with self._connect() as conn:
                conn.execute("PRAGMA journal_mode=WAL")
                conn.execute(
                    "CREATE TABLE pending_signals (id INTEGER PRIMARY KEY, signal_json TEXT NOT NULL, created_at REAL NOT NULL, retry_count INTEGER DEFAULT 0, sent_at REAL DEFAULT NULL)"
                )
                conn.execute(
                    "CREATE TABLE community_signatures (signature_id TEXT PRIMARY KEY, signature_json TEXT NOT NULL, received_at REAL NOT NULL, expires_at REAL DEFAULT NULL)"
                )
                conn.execute("CREATE TABLE metadata (key TEXT PRIMARY KEY, value TEXT NOT NULL)")
                conn.commit()

    def _cleanup_old_signals(self) -> None:
        if not self.enabled:
            return
        now = time.time()
        with self._lock:
            with self._connect() as conn:
                conn.execute("DELETE FROM pending_signals WHERE sent_at IS NOT NULL AND sent_at < ?", (now - 86400.0,))
                conn.execute("DELETE FROM pending_signals WHERE sent_at IS NULL AND created_at < ?", (now - 7 * 86400.0,))
                conn.execute(
                    "DELETE FROM community_signatures WHERE expires_at IS NOT NULL AND expires_at > 0 AND expires_at < ?",
                    (now,),
                )
                conn.commit()

    def get_stats(self) -> CommunityStats:
        if not self.enabled:
            return CommunityStats(0.0, 0, 0, 0, 0, 0.0, "offline", 0.0, 0)
        with self._lock:
            with self._connect() as conn:
                pending_row = conn.execute("SELECT COUNT(*) AS c FROM pending_signals WHERE sent_at IS NULL").fetchone()
                cached_row = conn.execute("SELECT COUNT(*) AS c FROM community_signatures").fetchone()
        pending = int(pending_row["c"]) if pending_row else 0
        cached = int(cached_row["c"]) if cached_row else 0
        received = int(self._signatures_received)
        ratio = (self._signals_sent / float(max(1, received))) if received > 0 else float(self._signals_sent)
        return CommunityStats(
            trust_score=float(self._trust_score),
            signals_sent=int(self._signals_sent),
            signals_pending=pending,
            signatures_received=received,
            signatures_cached=cached,
            last_sync=float(self._last_sync),
            hub_status=str(self._hub_status),
            contribution_ratio=float(ratio),
            community_size=int(self._community_size),
        )

    def get_status(self) -> dict[str, Any]:
        stats = self.get_stats()
        return {
            "enabled": bool(self.enabled),
            "privacy_level": int(self.privacy_level),
            "hub_status": stats.hub_status,
            "signals_pending": int(stats.signals_pending),
            "signatures_cached": int(stats.signatures_cached),
            "trust_score": float(stats.trust_score),
            "last_sync": float(stats.last_sync) if stats.last_sync > 0 else None,
        }



from __future__ import annotations

import json
from pathlib import Path

import httpx
import pytest

from orchesis.api import create_api_app
from orchesis.reliability import ReliabilityReportGenerator


def test_reliability_report_generates(tmp_path: Path) -> None:
    generator = ReliabilityReportGenerator(
        corpus_path=str(Path(__file__).parent / "corpus"),
        decisions_log=str(tmp_path / "decisions.jsonl"),
        fuzz_log=str(tmp_path / "fuzz_runs.jsonl"),
        mutation_log=str(tmp_path / "mutation_runs.jsonl"),
        replay_log=str(tmp_path / "replay_runs.jsonl"),
    )
    report = generator.generate()
    assert report.generated_at
    assert report.orchesis_version
    assert report.total_tests >= 1
    assert report.corpus_entries >= 14


def test_reliability_markdown_format(tmp_path: Path) -> None:
    generator = ReliabilityReportGenerator(
        corpus_path=str(Path(__file__).parent / "corpus"),
        decisions_log=str(tmp_path / "decisions.jsonl"),
    )
    markdown = generator.to_markdown(generator.generate())
    assert "# Orchesis Reliability Report" in markdown
    assert "## Testing" in markdown
    assert "## Fuzzing" in markdown
    assert "## Runtime Guarantees" in markdown


def test_reliability_json_format(tmp_path: Path) -> None:
    generator = ReliabilityReportGenerator(
        corpus_path=str(Path(__file__).parent / "corpus"),
        decisions_log=str(tmp_path / "decisions.jsonl"),
    )
    payload = generator.to_json(generator.generate())
    parsed = json.loads(payload)
    assert "generated_at" in parsed
    assert "orchesis_version" in parsed
    assert "total_tests" in parsed


def test_reliability_prefers_fuzz_meta(tmp_path: Path) -> None:
    fuzz_meta = tmp_path / "fuzz_meta.json"
    fuzz_meta.write_text(
        json.dumps(
            {
                "total_runs": 14,
                "total_requests_lifetime": 140000,
                "total_mutations_lifetime": 14000,
                "total_bypasses_lifetime": 0,
                "total_invariant_checks": 126,
                "invariant_failures": 0,
                "first_run": "2026-02-28T03:00:00Z",
                "last_run": "2026-03-14T03:00:00Z",
                "consecutive_clean_runs": 14,
                "days_without_bypass": 14,
            }
        ),
        encoding="utf-8",
    )
    generator = ReliabilityReportGenerator(
        corpus_path=str(Path(__file__).parent / "corpus"),
        decisions_log=str(tmp_path / "decisions.jsonl"),
        fuzz_log=str(tmp_path / "fuzz_runs.jsonl"),
        mutation_log=str(tmp_path / "mutation_runs.jsonl"),
        replay_log=str(tmp_path / "replay_runs.jsonl"),
        fuzz_meta_path=str(fuzz_meta),
    )
    report = generator.generate()
    assert report.total_fuzz_requests == 140000
    assert report.total_mutations_tested == 14000
    assert report.total_invariant_checks == 126
    assert report.invariant_failures == 0
    assert report.consecutive_clean_runs == 14
    assert report.days_without_bypass == 14


@pytest.mark.asyncio
async def test_reliability_api_endpoint(tmp_path: Path) -> None:
    policy_path = tmp_path / "policy.yaml"
    policy_path.write_text(
        """
api:
  token: "orch_sk_test"
rules:
  - name: budget_limit
    max_cost_per_call: 1.0
""".strip(),
        encoding="utf-8",
    )
    app = create_api_app(
        policy_path=str(policy_path),
        state_persist=str(tmp_path / "state.jsonl"),
        decisions_log=str(tmp_path / "decisions.jsonl"),
        history_path=str(tmp_path / "policy_versions.jsonl"),
    )
    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        response = await client.get(
            "/api/v1/reliability",
            headers={"Authorization": "Bearer orch_sk_test"},
        )
    assert response.status_code == 200
    payload = response.json()
    assert "generated_at" in payload
    assert "orchesis_version" in payload



"""Plugin: Allow tool calls only during specific hours."""

from __future__ import annotations

from datetime import datetime, timezone
from zoneinfo import ZoneInfo

from orchesis.plugins import PluginInfo


class TimeWindowHandler:
    def _now(self, tz_name: str) -> datetime:
        zone = ZoneInfo(tz_name)
        return datetime.now(timezone.utc).astimezone(zone)

    def _parse_minutes(self, value: str) -> int:
        hour_text, minute_text = value.split(":", 1)
        hour = int(hour_text)
        minute = int(minute_text)
        if hour < 0 or hour > 23 or minute < 0 or minute > 59:
            raise ValueError("invalid time range")
        return hour * 60 + minute

    def evaluate(self, rule, request, **kwargs):  # noqa: ANN001, ANN003
        _ = (request, kwargs)
        checked = ["time_window"]
        window = rule.get("allowed_hours")
        if not isinstance(window, dict):
            return ["time_window: allowed_hours is required"], checked
        start = window.get("start")
        end = window.get("end")
        tz_name = window.get("timezone", "UTC")
        if not isinstance(start, str) or not isinstance(end, str) or not isinstance(tz_name, str):
            return ["time_window: allowed_hours requires start/end/timezone"], checked
        try:
            current = self._now(tz_name)
            now_minutes = current.hour * 60 + current.minute
            start_minutes = self._parse_minutes(start)
            end_minutes = self._parse_minutes(end)
        except Exception:
            return ["time_window: invalid allowed_hours configuration"], checked
        if start_minutes <= end_minutes:
            in_window = start_minutes <= now_minutes <= end_minutes
        else:
            in_window = now_minutes >= start_minutes or now_minutes <= end_minutes
        if not in_window:
            return [f"time_window: outside allowed window {start}-{end} {tz_name}"], checked
        return [], checked


PLUGIN_INFO = PluginInfo(
    name="time_window",
    rule_type="time_window",
    version="1.0",
    description="Allow tool calls only during specific hours",
    handler=TimeWindowHandler(),
)
